@@ -1,20 +1,21 @@
 import importlib
+import io
 import os
+import shutil
 import subprocess
 import sys
 import tempfile
-from configparser import ConfigParser
 from pathlib import Path
 from typing import (
     TYPE_CHECKING,
     Any,
-    Callable,
     Dict,
     List,
+    Literal,
     Optional,
+    TextIO,
     Tuple,
     Union,
-    no_type_check,
 )
 
 import py
@@ -27,19 +28,22 @@ from mypy import build
 from mypy.fscache import FileSystemCache
 from mypy.main import process_options
 
-if TYPE_CHECKING:
-    from _pytest._code.code import _TracebackStyle
-
-from pytest_mypy_plugins import utils
+from pytest_mypy_plugins import configs, utils
 from pytest_mypy_plugins.collect import File, YamlTestFile
 from pytest_mypy_plugins.utils import (
     OutputMatcher,
     TypecheckAssertionError,
     assert_expected_matched_actual,
-    capture_std_streams,
     fname_to_module,
 )
 
+if TYPE_CHECKING:
+    # pytest 8.3.0 renamed _TracebackStyle to TracebackStyle, but there is no syntax
+    # to assert what version you have using static conditions, so it has to be
+    # manually re-defined here. Once minimum supported pytest version is >= 8.3.0,
+    # the following can be replaced with `from _pytest._code.code import TracebackStyle`
+    TracebackStyle = Literal["long", "short", "line", "no", "native", "value", "auto"]
+
 
 class TraceLastReprEntry(ReprEntry):
     def toterminal(self, tw: TerminalWriter) -> None:
@@ -87,15 +91,19 @@ class ReturnCodes:
     FATAL_ERROR = 2
 
 
-def run_mypy_typechecking(cmd_options: List[str]) -> int:
+def run_mypy_typechecking(cmd_options: List[str], stdout: TextIO, stderr: TextIO) -> int:
     fscache = FileSystemCache()
     sources, options = process_options(cmd_options, fscache=fscache)
 
     error_messages = []
 
-    def flush_errors(new_messages: List[str], serious: bool) -> None:
+    # Different mypy versions have different arity of `flush_errors`: 2 and 3 params
+    def flush_errors(*args: Any) -> None:
+        new_messages: List[str]
+        serious: bool
+        *_, new_messages, serious = args
         error_messages.extend(new_messages)
-        f = sys.stderr if serious else sys.stdout
+        f = stderr if serious else stdout
         try:
             for msg in new_messages:
                 f.write(msg + "\n")
@@ -104,10 +112,19 @@ def run_mypy_typechecking(cmd_options: List[str]) -> int:
             sys.exit(ReturnCodes.FATAL_ERROR)
 
     try:
-        build.build(sources, options, flush_errors=flush_errors, fscache=fscache)
+        build.build(sources, options, flush_errors=flush_errors, fscache=fscache, stdout=stdout, stderr=stderr)
 
     except SystemExit as sysexit:
-        return sysexit.code
+        # The code to a SystemExit is optional
+        # From python docs, if the code is None then the exit code is 0
+        # Otherwise if the code is not an integer the exit code is 1
+        code = sysexit.code
+        if code is None:
+            code = 0
+        elif not isinstance(code, int):
+            code = 1
+
+        return code
     finally:
         fscache.flush()
 
@@ -117,6 +134,184 @@ def run_mypy_typechecking(cmd_options: List[str]) -> int:
     return ReturnCodes.SUCCESS
 
 
+class MypyExecutor:
+    def __init__(
+        self,
+        same_process: bool,
+        rootdir: Union[Path, None],
+        execution_path: Path,
+        environment_variables: Dict[str, Any],
+        mypy_executable: str,
+    ) -> None:
+        self.rootdir = rootdir
+        self.same_process = same_process
+        self.execution_path = execution_path
+        self.mypy_executable = mypy_executable
+        self.environment_variables = environment_variables
+
+    def execute(self, mypy_cmd_options: List[str]) -> Tuple[int, Tuple[str, str]]:
+        # Returns (returncode, (stdout, stderr))
+        if self.same_process:
+            return self._typecheck_in_same_process(mypy_cmd_options)
+        else:
+            return self._typecheck_in_new_subprocess(mypy_cmd_options)
+
+    def _typecheck_in_new_subprocess(self, mypy_cmd_options: List[Any]) -> Tuple[int, Tuple[str, str]]:
+        # add current directory to path
+        self._collect_python_path(self.rootdir)
+        # adding proper MYPYPATH variable
+        self._collect_mypy_path(self.rootdir)
+
+        # Windows requires this to be set, otherwise the interpreter crashes
+        if "SYSTEMROOT" in os.environ:
+            self.environment_variables["SYSTEMROOT"] = os.environ["SYSTEMROOT"]
+
+        completed = subprocess.run(
+            [self.mypy_executable, *mypy_cmd_options],
+            capture_output=True,
+            cwd=os.getcwd(),
+            env=self.environment_variables,
+        )
+        captured_stdout = completed.stdout.decode()
+        captured_stderr = completed.stderr.decode()
+        return completed.returncode, (captured_stdout, captured_stderr)
+
+    def _typecheck_in_same_process(self, mypy_cmd_options: List[Any]) -> Tuple[int, Tuple[str, str]]:
+        return_code = -1
+        with utils.temp_environ(), utils.temp_path(), utils.temp_sys_modules():
+            # add custom environment variables
+            for key, val in self.environment_variables.items():
+                os.environ[key] = val
+
+            # add current directory to path
+            sys.path.insert(0, str(self.execution_path))
+
+            stdout = io.StringIO()
+            stderr = io.StringIO()
+
+            with stdout, stderr:
+                return_code = run_mypy_typechecking(mypy_cmd_options, stdout=stdout, stderr=stderr)
+                stdout_value = stdout.getvalue()
+                stderr_value = stderr.getvalue()
+
+            return return_code, (stdout_value, stderr_value)
+
+    def _collect_python_path(self, rootdir: Optional[Path]) -> None:
+        python_path_parts = []
+
+        existing_python_path = os.environ.get("PYTHONPATH")
+        if existing_python_path:
+            python_path_parts.append(existing_python_path)
+        python_path_parts.append(str(self.execution_path))
+        python_path_key = self.environment_variables.get("PYTHONPATH")
+        if python_path_key:
+            python_path_parts.append(maybe_to_abspath(python_path_key, rootdir))
+            python_path_parts.append(python_path_key)
+
+        self.environment_variables["PYTHONPATH"] = ":".join(python_path_parts)
+
+    def _collect_mypy_path(self, rootdir: Optional[Path]) -> None:
+        mypy_path_parts = []
+
+        existing_mypy_path = os.environ.get("MYPYPATH")
+        if existing_mypy_path:
+            mypy_path_parts.append(existing_mypy_path)
+        mypy_path_key = self.environment_variables.get("MYPYPATH")
+        if mypy_path_key:
+            mypy_path_parts.append(maybe_to_abspath(mypy_path_key, rootdir))
+            mypy_path_parts.append(mypy_path_key)
+        if rootdir:
+            mypy_path_parts.append(str(rootdir))
+
+        self.environment_variables["MYPYPATH"] = ":".join(mypy_path_parts)
+
+
+class OutputChecker:
+    def __init__(self, expect_fail: bool, execution_path: Path, expected_output: List[OutputMatcher]) -> None:
+        self.expect_fail = expect_fail
+        self.execution_path = execution_path
+        self.expected_output = expected_output
+
+    def check(self, ret_code: int, stdout: str, stderr: str) -> None:
+        mypy_output = stdout + stderr
+        if ret_code == ReturnCodes.FATAL_ERROR:
+            print(mypy_output, file=sys.stderr)
+            raise TypecheckAssertionError(error_message="Critical error occurred")
+
+        output_lines = []
+        for line in mypy_output.splitlines():
+            output_line = replace_fpath_with_module_name(line, rootdir=self.execution_path)
+            output_lines.append(output_line)
+        try:
+            assert_expected_matched_actual(expected=self.expected_output, actual=output_lines)
+        except TypecheckAssertionError as e:
+            if not self.expect_fail:
+                raise e
+        else:
+            if self.expect_fail:
+                raise TypecheckAssertionError("Expected failure, but test passed")
+
+
+class Runner:
+    def __init__(
+        self,
+        *,
+        files: List[File],
+        config: Config,
+        main_file: Path,
+        config_file: Optional[str],
+        disable_cache: bool,
+        mypy_executor: MypyExecutor,
+        output_checker: OutputChecker,
+        test_only_local_stub: bool,
+        incremental_cache_dir: str,
+    ) -> None:
+        self.files = files
+        self.config = config
+        self.main_file = main_file
+        self.config_file = config_file
+        self.mypy_executor = mypy_executor
+        self.disable_cache = disable_cache
+        self.output_checker = output_checker
+        self.test_only_local_stub = test_only_local_stub
+        self.incremental_cache_dir = incremental_cache_dir
+
+    def run(self) -> None:
+        # start from main.py
+        mypy_cmd_options = self._prepare_mypy_cmd_options()
+        mypy_cmd_options.append(str(self.main_file))
+
+        # make files
+        for file in self.files:
+            self._make_test_file(file)
+
+        returncode, (stdout, stderr) = self.mypy_executor.execute(mypy_cmd_options)
+        self.output_checker.check(returncode, stdout, stderr)
+
+    def _make_test_file(self, file: File) -> None:
+        current_directory = Path.cwd()
+        fpath = current_directory / file.path
+        fpath.parent.mkdir(parents=True, exist_ok=True)
+        fpath.write_text(file.content)
+
+    def _prepare_mypy_cmd_options(self) -> List[str]:
+        mypy_cmd_options = [
+            "--show-traceback",
+            "--no-error-summary",
+            "--no-pretty",
+            "--hide-error-context",
+        ]
+        if not self.test_only_local_stub:
+            mypy_cmd_options.append("--no-silence-site-packages")
+        if not self.disable_cache:
+            mypy_cmd_options.extend(["--cache-dir", self.incremental_cache_dir])
+
+        if self.config_file:
+            mypy_cmd_options.append(f"--config-file={self.config_file}")
+
+        return mypy_cmd_options
+
+
 class YamlTestItem(pytest.Item):
     def __init__(
         self,
@@ -147,22 +342,21 @@ class YamlTestItem(pytest.Item):
 
         # config parameters
         self.root_directory = self.config.option.mypy_testing_base
+
+        # You cannot use both `.ini` and `pyproject.toml` files at the same time:
+        if self.config.option.mypy_ini_file and self.config.option.mypy_pyproject_toml_file:
+            raise ValueError("Cannot specify both `--mypy-ini-file` and `--mypy-pyproject-toml-file`")
+
         if self.config.option.mypy_ini_file:
             self.base_ini_fpath = os.path.abspath(self.config.option.mypy_ini_file)
         else:
             self.base_ini_fpath = None
+        if self.config.option.mypy_pyproject_toml_file:
+            self.base_pyproject_toml_fpath = os.path.abspath(self.config.option.mypy_pyproject_toml_file)
+        else:
+            self.base_pyproject_toml_fpath = None
         self.incremental_cache_dir = os.path.join(self.root_directory, ".mypy_cache")
 
-    def make_test_file(self, file: File) -> None:
-        current_directory = Path.cwd()
-        fpath = current_directory / file.path
-        fpath.parent.mkdir(parents=True, exist_ok=True)
-        fpath.write_text(file.content)
-
-    def make_test_files_in_current_directory(self) -> None:
-        for file in self.files:
-            self.make_test_file(file)
-
     def remove_cache_files(self, fpath_no_suffix: Path) -> None:
         cache_file = Path(self.incremental_cache_dir)
         cache_file /= ".".join([str(part) for part in sys.version_info[:2]])
@@ -184,51 +378,6 @@ class YamlTestItem(pytest.Item):
             ):
                 parent_dir.rmdir()
 
-    def typecheck_in_new_subprocess(
-        self, execution_path: Path, mypy_cmd_options: List[Any]
-    ) -> Tuple[int, Tuple[str, str]]:
-        import shutil
-
-        mypy_executable = shutil.which("mypy")
-        assert mypy_executable is not None, "mypy executable is not found"
-
-        rootdir = getattr(getattr(self.parent, "config", None), "rootdir", None)
-        # add current directory to path
-        self._collect_python_path(rootdir, execution_path)
-        # adding proper MYPYPATH variable
-        self._collect_mypy_path(rootdir)
-
-        # Windows requires this to be set, otherwise the interpreter crashes
-        if "SYSTEMROOT" in os.environ:
-            self.environment_variables["SYSTEMROOT"] = os.environ["SYSTEMROOT"]
-
-        completed = subprocess.run(
-            [mypy_executable, *mypy_cmd_options],
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
-            cwd=os.getcwd(),
-            env=self.environment_variables,
-        )
-        captured_stdout = completed.stdout.decode()
-        captured_stderr = completed.stderr.decode()
-        return completed.returncode, (captured_stdout, captured_stderr)
-
-    def typecheck_in_same_process(
-        self, execution_path: Path, mypy_cmd_options: List[Any]
-    ) -> Tuple[int, Tuple[str, str]]:
-        with utils.temp_environ(), utils.temp_path(), utils.temp_sys_modules():
-            # add custom environment variables
-            for key, val in self.environment_variables.items():
-                os.environ[key] = val
-
-            # add current directory to path
-            sys.path.insert(0, str(execution_path))
-
-            with capture_std_streams() as (stdout, stderr):
-                return_code = run_mypy_typechecking(mypy_cmd_options)
-
-            return return_code, (stdout.getvalue(), stderr.getvalue())
-
     def execute_extension_hook(self) -> None:
         extension_hook_fqname = self.config.option.mypy_extension_hook
         module_name, func_name = extension_hook_fqname.rsplit(".", maxsplit=1)
@@ -241,53 +390,47 @@ class YamlTestItem(pytest.Item):
             temp_dir = tempfile.TemporaryDirectory(prefix="pytest-mypy-", dir=self.root_directory)
 
         except (FileNotFoundError, PermissionError, NotADirectoryError) as e:
-
             raise TypecheckAssertionError(
                 error_message=f"Testing base directory {self.root_directory} must exist and be writable"
             ) from e
 
         try:
+            mypy_executable = shutil.which("mypy")
+            assert mypy_executable is not None, "mypy executable is not found"
+            rootdir = getattr(getattr(self.parent, "config", None), "rootdir", None)
+
+            # extension point for derived packages
+            if (
+                hasattr(self.config.option, "mypy_extension_hook")
+                and self.config.option.mypy_extension_hook is not None
+            ):
+                self.execute_extension_hook()
+
             execution_path = Path(temp_dir.name)
-
             with utils.cd(execution_path):
-                # extension point for derived packages
-                if (
-                    hasattr(self.config.option, "mypy_extension_hook")
-                    and self.config.option.mypy_extension_hook is not None
-                ):
-                    self.execute_extension_hook()
+                mypy_executor = MypyExecutor(
+                    same_process=self.same_process,
+                    execution_path=execution_path,
+                    rootdir=rootdir,
+                    environment_variables=self.environment_variables,
+                    mypy_executable=mypy_executable,
+                )
 
-                # start from main.py
-                main_file = str(execution_path / "main.py")
-                mypy_cmd_options = self.prepare_mypy_cmd_options(execution_path)
-                mypy_cmd_options.append(main_file)
-
-                # make files
-                self.make_test_files_in_current_directory()
-
-                if self.same_process:
-                    returncode, (stdout, stderr) = self.typecheck_in_same_process(execution_path, mypy_cmd_options)
-                else:
-                    returncode, (stdout, stderr) = self.typecheck_in_new_subprocess(execution_path, mypy_cmd_options)
-
-                mypy_output = stdout + stderr
-                if returncode == ReturnCodes.FATAL_ERROR:
-                    print(mypy_output, file=sys.stderr)
-                    raise TypecheckAssertionError(error_message="Critical error occurred")
-
-                output_lines = []
-                for line in mypy_output.splitlines():
-                    output_line = replace_fpath_with_module_name(line, rootdir=execution_path)
-                    output_lines.append(output_line)
-                try:
-                    assert_expected_matched_actual(expected=self.expected_output, actual=output_lines)
-                except TypecheckAssertionError as e:
-                    if not self.expect_fail:
-                        raise e
-                else:
-                    if self.expect_fail:
-                        raise TypecheckAssertionError("Expected failure, but test passed")
+                output_checker = OutputChecker(
+                    expect_fail=self.expect_fail, execution_path=execution_path, expected_output=self.expected_output
+                )
 
+                Runner(
+                    files=self.files,
+                    config=self.config,
+                    main_file=execution_path / "main.py",
+                    config_file=self.prepare_config_file(execution_path),
+                    disable_cache=self.disable_cache,
+                    mypy_executor=mypy_executor,
+                    output_checker=output_checker,
+                    test_only_local_stub=self.test_only_local_stub,
+                    incremental_cache_dir=self.incremental_cache_dir,
+                ).run()
         finally:
             temp_dir.cleanup()
             # remove created modules
@@ -298,39 +441,23 @@ class YamlTestItem(pytest.Item):
 
         assert not os.path.exists(temp_dir.name)
 
-    def prepare_mypy_cmd_options(self, execution_path: Path) -> List[str]:
-        mypy_cmd_options = [
-            "--show-traceback",
-            "--no-error-summary",
-            "--no-pretty",
-            "--hide-error-context",
-        ]
-        if not self.test_only_local_stub:
-            mypy_cmd_options.append("--no-silence-site-packages")
-        if not self.disable_cache:
-            mypy_cmd_options.extend(["--cache-dir", self.incremental_cache_dir])
-
-        # Merge `self.base_ini_fpath` and `self.additional_mypy_config`
+    def prepare_config_file(self, execution_path: Path) -> Optional[str]:
+        # Merge (`self.base_ini_fpath` or `base_pyproject_toml_fpath`)
+        # and `self.additional_mypy_config`
         # into one file and copy to the typechecking folder:
-        mypy_ini_config = ConfigParser()
-        if self.base_ini_fpath:
-            mypy_ini_config.read(self.base_ini_fpath)
-        if self.additional_mypy_config:
-            additional_config = self.additional_mypy_config
-            if "[mypy]" not in additional_config:
-                additional_config = "[mypy]\n" + additional_config
-            mypy_ini_config.read_string(additional_config)
-
-        if mypy_ini_config.sections():
-            mypy_config_file_path = execution_path / "mypy.ini"
-            with mypy_config_file_path.open("w") as f:
-                mypy_ini_config.write(f)
-            mypy_cmd_options.append(f"--config-file={str(mypy_config_file_path)}")
-
-        return mypy_cmd_options
+        if self.base_pyproject_toml_fpath:
+            return configs.join_toml_configs(
+                self.base_pyproject_toml_fpath, self.additional_mypy_config, execution_path
+            )
+        elif self.base_ini_fpath or self.additional_mypy_config:
+            # We might have `self.base_ini_fpath` set as well.
+            # Or this might be a legacy case: only `mypy_config:` is set in the `yaml` test case.
+            # This means that no real file is provided.
+            return configs.join_ini_configs(self.base_ini_fpath, self.additional_mypy_config, execution_path)
+        return None
 
     def repr_failure(
-        self, excinfo: ExceptionInfo[BaseException], style: Optional["_TracebackStyle"] = None
+        self, excinfo: ExceptionInfo[BaseException], style: Optional["TracebackStyle"] = None
     ) -> Union[str, TerminalRepr]:
         if excinfo.errisinstance(SystemExit):
             # We assume that before doing exit() (which raises SystemExit) we've printed
@@ -345,7 +472,7 @@ class YamlTestItem(pytest.Item):
             repr_file_location = ReprFileLocation(
                 path=self.fspath, lineno=self.starting_lineno + excinfo.value.lineno, message=""  # type: ignore
             )
-            repr_tb_entry = TraceLastReprEntry(
+            repr_tb_entry = ReprEntry(
                 exception_repr.reprtraceback.reprentries[-1].lines[1:], None, None, repr_file_location, "short"
             )
             exception_repr.reprtraceback.reprentries = [repr_tb_entry]
@@ -353,40 +480,5 @@ class YamlTestItem(pytest.Item):
         else:
             return super().repr_failure(excinfo, style="native")
 
-    @no_type_check
-    def reportinfo(self) -> Tuple[Union[py.path.local, Path, str], Optional[int], str]:
-        # To support both Pytest 6.x and 7.x
-        path = getattr(self, "path", None) or getattr(self, "fspath")
-        return path, None, self.name
-
-    def _collect_python_path(
-        self,
-        rootdir: Optional[Path],
-        execution_path: Path,
-    ) -> None:
-        python_path_parts = []
-
-        existing_python_path = os.environ.get("PYTHONPATH")
-        if existing_python_path:
-            python_path_parts.append(existing_python_path)
-        if execution_path:
-            python_path_parts.append(str(execution_path))
-        python_path_key = self.environment_variables.get("PYTHONPATH")
-        if python_path_key:
-            python_path_parts.append(maybe_to_abspath(python_path_key, rootdir))
-            python_path_parts.append(python_path_key)
-
-        self.environment_variables["PYTHONPATH"] = ":".join(python_path_parts)
-
-    def _collect_mypy_path(self, rootdir: Optional[Path]) -> None:
-        mypy_path_parts = []
-
-        existing_mypy_path = os.environ.get("MYPYPATH")
-        if existing_mypy_path:
-            mypy_path_parts.append(existing_mypy_path)
-        mypy_path_key = self.environment_variables.get("MYPYPATH")
-        if mypy_path_key:
-            mypy_path_parts.append(maybe_to_abspath(mypy_path_key, rootdir))
-            mypy_path_parts.append(mypy_path_key)
-
-        self.environment_variables["MYPYPATH"] = ":".join(mypy_path_parts)
+    def reportinfo(self) -> Tuple[Union[Path, str], Optional[int], str]:
+        return self.path, None, self.name
